#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#include <X11/xpm.h>

#include "hruler.xpm"
#include "vruler.xpm"

extern char *optarg;
extern int optind, opterr, optopt;
int getopt(int, char * const [], const char *);


void
usage() {
    fprintf(stderr, "usage: xruler [-display name] [-fg color] [-bg color] [-width number] [-height number]\n");
   exit(0);
}
int
main(int argc, char *argv[])
{
    int RULER_HEIGHT = 100;
    int RULER_WIDTH = 40;
    Display *display;
    Screen *screen;
    XSizeHints  size_hints;
    XWMHints    wm_hints;
    XClassHint  class_hints;
    Pixmap pixmap, mask, realmask;
    Window win;
    XpmAttributes attr;
    GC gc, maskgc;
    XGCValues           gc_values;
    int c, height = 400, width = 400;
    enum { HORIZONTAL, VERTICAL } mode = VERTICAL;
    const char *fgcolor = "white", *bgcolor = "black";
    XColor bgc, fgc, tmpcolor;
    char *disp = NULL;
    
    c = 1;
    while (c < argc-1) {
	if (!strcasecmp(argv[c], "-height")) {
	    height = atoi(argv[++c]);
	    RULER_HEIGHT = 100;
	    RULER_WIDTH = 40;
	    mode = VERTICAL;
	    if (height == 0)
		usage();
	    c++;
	} else if (!strcasecmp(argv[c], "-width")) {
	    width = atoi(argv[++c]);
	    RULER_HEIGHT = 40;
	    RULER_WIDTH = 100;
	    mode = HORIZONTAL;
	    if (width == 0)
		usage();
	    c++;
	} else if (!strcasecmp(argv[c], "-fg") ||
		   !strcasecmp(argv[c], "-foreground")) {
	    fgcolor = argv[++c];
	    c++;
	} else if (!strcasecmp(argv[c], "-d") ||
		   !strcasecmp(argv[c], "-display")) {
	    disp = argv[++c];
	    c++;
	} else if (!strcasecmp(argv[c], "-bg") ||
		   !strcasecmp(argv[c], "-background")) {
	    bgcolor = argv[++c];
	    c++;
	} else {
	    usage();
	}
    }
    if (c != argc)
	usage();

    if ((display = XOpenDisplay(disp)) == NULL) {
	fprintf(stderr, "Cannot open local X-display.\n");
	exit(1);
    }
    screen = DefaultScreenOfDisplay(display);
    
    if (XShapeQueryExtension(display, &c, &c) == False) {
	fprintf(stderr, "Could not find X-Shape-Extension.\n");
	exit(1);
    }


    // allocate colors
    XAllocNamedColor(display, DefaultColormapOfScreen(screen),
		     fgcolor, &fgc, &tmpcolor);
    XAllocNamedColor(display, DefaultColormapOfScreen(screen),
		     bgcolor, &bgc, &tmpcolor);

    attr.valuemask = XpmDepth|XpmColorSymbols;
    attr.depth = DefaultDepthOfScreen(screen);
    attr.colorsymbols = malloc(sizeof(XpmColorSymbol)*2);
    attr.numsymbols = 2;
    attr.colorsymbols[0].name = "fg";
    attr.colorsymbols[0].pixel = fgc.pixel;
    attr.colorsymbols[1].name = "bg";
    attr.colorsymbols[1].pixel = bgc.pixel;

    gc = DefaultGC(display, DefaultScreen(display));

    // setup some hints to the window manager (mainly the class/name and the
    // ability to not resize the window)
    size_hints.flags = PSize | PMinSize | PMaxSize;
    if (mode == VERTICAL) {
	win = XCreateSimpleWindow(display, DefaultRootWindow(display),
				  0, 0, RULER_WIDTH, height, 2,
				  fgc.pixel, bgc.pixel);
	size_hints.width = size_hints.min_width = size_hints.max_width = RULER_WIDTH;
	size_hints.height = size_hints.min_height = size_hints.max_height = height;
	width = RULER_WIDTH;
    } else {
	win = XCreateSimpleWindow(display, DefaultRootWindow(display),
				  0, 0, width, RULER_HEIGHT, 2,
				  fgc.pixel, bgc.pixel);
	size_hints.width = size_hints.min_width = size_hints.max_width = width;
	size_hints.height = size_hints.min_height = size_hints.max_height = RULER_HEIGHT;
	height = RULER_HEIGHT;
    }

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = "xruler";
    class_hints.res_class = "XRuler";

    XmbSetWMProperties(display, win, "XRuler", "XRuler",
		       argv, argc, &size_hints, &wm_hints, &class_hints);


    // load pixmap
    if (mode == VERTICAL)
	XpmCreatePixmapFromData(display, win, vruler_xpm, &pixmap, &mask, &attr);
    else
	XpmCreatePixmapFromData(display, win, hruler_xpm, &pixmap, &mask, &attr);

    // create the real mask for the shaped window. we have a mask from the
    // pixmap, now we need to copy it over and over until we fill up the height
    // of the real window
    realmask = XCreatePixmap(display, win, width, height, 1);
    maskgc = XCreateGC(display, mask, 0, &gc_values);
    if (mode == VERTICAL) {
	for (c = 0; c != height;) {
	    int h;
	    if (height-c < RULER_HEIGHT -1) {
		h = height - c;
	    } else {
		h = RULER_HEIGHT;
	    }
	    XCopyArea(display, mask, realmask, maskgc, 0, 0,
		      RULER_WIDTH, h, 0, c);
	    c += h;
	}
    } else {
	for (c = 0; c != width;) {
	    int h;
	    if (width-c < RULER_WIDTH -1) {
		h = width - c;
	    } else {
		h = RULER_WIDTH;
	    }
	    XCopyArea(display, mask, realmask, maskgc, 0, 0,
		      h, RULER_HEIGHT, c, 0);
	    c += h;
	}
    }

    // reshape!
    XShapeCombineMask(display, win, ShapeBounding, 0, 0, realmask, ShapeSet);

    // get expose events and button press events from the window
    XSelectInput(display, win, ExposureMask|ButtonPressMask);

    XFlush(display);
    XMapWindow(display, win);

    for(;;) {
	XEvent ev;
	XNextEvent(display, &ev);
	if (ev.type == Expose) {
	    int drawn = 0;
	    if (mode == VERTICAL) {
		while (drawn != height) {
		    int h;
		    if (height-drawn < RULER_HEIGHT) {
			h = height - drawn;
		    } else {
			h = RULER_HEIGHT;
		    }
		    XCopyArea(display, pixmap, win, gc, 0, 0,
			      RULER_WIDTH, h, 0, drawn);
		    drawn += h;
		}
	    } else {
		while (drawn != width) {
		    int h;
		    if (width-drawn < RULER_WIDTH) {
			h = width - drawn;
		    } else {
			h = RULER_WIDTH;
		    }
		    XCopyArea(display, pixmap, win, gc, 0, 0,
			      h, RULER_HEIGHT, drawn, 0);
		    drawn += h;
		}
	    }
	} else if (ev.type == ButtonPress) {
	    return 0;
	}
    }
}

